@@ -5,9 +5,9 @@ from __future__ import annotations
 Modern Robotics: Mechanics, Planning, and Control.
 Code Library
 ***************************************************************************
-Author: Huan Weng, Bill Hunt, Jarvis Schultz, Mikhail Todes, Minchang Sung
-Email: tjdalsckd@gmail.com
-Date: January 2023
+Author: Huan Weng, Bill Hunt, Jarvis Schultz, Mikhail Todes,
+Email: huanweng@u.northwestern.edu
+Date: January 2018
 ***************************************************************************
 Language: Python
 Also available in: MATLAB, Mathematica
@@ -693,6 +693,19 @@ def DerivativeJacobianSpace(Slist,thetalist):
                     dvecJs[6*j:6*j+6,i] = ad(Js_j)*Js_i              
     return dvecJs    
 def AnalyticJacobianBody(M,Blist, thetalist):
+    Jb =  JacobianBody(Blist, thetalist)
+    Tsb = FKinBody(M,Blist, thetalist)
+    Rsb= Tsb[0:3,0:3]
+    r = so3ToVec(Rsb)
+    norm_r = np.sqrt(r[0]*r[0]+r[1]*r[1]+r[2]*r[2])
+    omega_r = VecToso3(r)
+    A = np.eye(3) - (1-np.cos(norm_r))/(norm_r*norm_r) * omega_r+ (norm_r - np.sin(norm_r)) / (norm_r**3) * omega_r @ omega_r
+    A_ = np.eye(6)
+    A_[0:3,0:3] = A.T
+    A_[3:6,3:6] = Rsb
+    Ja = A_ @ Jb
+    return Ja
+def bulletAnalyticJacobianBody(M,Blist, thetalist):
     Jb =  JacobianBody(Blist, thetalist)
     Tsb = FKinBody(M,Blist, thetalist)
     Rsb= Tsb[0:3,0:3]
@@ -705,7 +718,21 @@ def AnalyticJacobianBody(M,Blist, thetalist):
     A_[3:6,3:6] = Rsb
     Ja = A_ @ Jb
     return Ja
+    
 def AnalyticJacobianSpace(M,Slist, thetalist):
+    Js =  JacobianSpace(Slist, thetalist)
+    Tsb = FKinSpace(M,Slist, thetalist)
+    Rsb= Tsb[0:3,0:3]
+    r = so3ToVec(Rsb)
+    norm_r = np.sqrt(r[0]*r[0]+r[1]*r[1]+r[2]*r[2])
+    omega_r = VecToso3(r)
+    A = np.eye(3) - (1-np.cos(norm_r))/(norm_r*norm_r) * omega_r+ (norm_r - np.sin(norm_r)) / (norm_r**3) * omega_r @ omega_r
+    A_ = np.eye(6)
+    A_[0:3,0:3] = A.T
+    A_[3:6,3:6] = Rsb
+    Ja = A_ @ Js
+    return Ja   
+def bulletAnalyticJacobianSpace(M,Slist, thetalist):
     Js =  JacobianSpace(Slist, thetalist)
     Tsb = FKinSpace(M,Slist, thetalist)
     Rsb= Tsb[0:3,0:3]
@@ -717,7 +744,7 @@ def AnalyticJacobianSpace(M,Slist, thetalist):
     A_[0:3,0:3] = Rsb
     A_[3:6,3:6] = Rsb
     Ja = A_ @ Js
-    return Ja    
+    return Ja       
 def JacobianSpace(Slist, thetalist):
     """Computes the space Jacobian for an open chain robot
 
@@ -1682,7 +1709,7 @@ def JointTrajectory(thetastart, thetaend, Tf, N, method):
     traj = np.array(traj).T
     return traj
 
-def ScrewTrajectory(Xstart, Xend, Tf, N, method):
+def ScrewTrajectory(Xstart, Xend, Tf, N, method,type=0):
     """Computes a trajectory as a list of N SE(3) matrices corresponding to
       the screw motion about a space screw axis
 
@@ -1734,17 +1761,26 @@ def ScrewTrajectory(Xstart, Xend, Tf, N, method):
     for i in range(N):
         if method == 3:
             s = CubicTimeScaling(Tf, timegap * i)
+            sdot = CubicTimeScalingDot(Tf, timegap * i)
+            sddot = CubicTimeScalingDdot(Tf, timegap * i)        
         else:
             s = QuinticTimeScaling(Tf, timegap * i)
-        traj[i] \
-        = np.dot(Xstart, MatrixExp6(MatrixLog6(np.dot(TransInv(Xstart), \
-                                                      Xend)) * s))
+            sdot = QuinticTimeScalingDot(Tf, timegap * i)
+            sddot = QuinticTimeScalingDdot(Tf, timegap * i)    
+        Xs = np.dot(Xstart, MatrixExp6(MatrixLog6(np.dot(TransInv(Xstart),Xend)) * s)) 
+        V = Adjoint(Xs)@se3ToVec(MatrixLog6(TransInv(Xstart)@Xend)); 
+        V = Adjoint(Xstart) @ se3ToVec(MatrixLog6(TransInv(Xstart) @ Xend));
+        if type == 0  :
+            traj[i] = Xs
+        elif type == 1 :
+            traj[i] = V*sdot
+        elif type == 2 :   
+            traj[i] = V*sddot
     return traj
 
 def CartesianTrajectory(Xstart, Xend, Tf, N, method,type=0):
     """Computes a trajectory as a list of N SE(3) matrices corresponding to
     the origin of the end-effector frame following a straight line
-    ** IMPROVE VERSION**
 
     :param Xstart: The initial end-effector configuration
     :param Xend: The final end-effector configuration
@@ -2072,3 +2108,4 @@ def SimulateControl(thetalist, dthetalist, g, Ftipmat, Mlist, Glist, \
     taumat = np.array(taumat).T
     thetamat = np.array(thetamat).T
     return (taumat, thetamat)
+
